package org.eclipse.jdt.internal.compiler.ast;

package .var.folders._j.vfbv187d0bb8b8w5spzgg8180000gn.T;

import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedHashSet;
import java.util.Set;
import org.eclipse.jdt.core.compiler.CategorizedProblem;
import org.eclipse.jdt.core.compiler.CharOperation;
import org.eclipse.jdt.internal.compiler.ASTVisitor;
import org.eclipse.jdt.internal.compiler.ClassFile;
import org.eclipse.jdt.internal.compiler.CompilationResult;
import org.eclipse.jdt.internal.compiler.IErrorHandlingPolicy;
import org.eclipse.jdt.internal.compiler.ast.ASTNode;
import org.eclipse.jdt.internal.compiler.ast.AbstractMethodDeclaration;
import org.eclipse.jdt.internal.compiler.ast.AbstractVariableDeclaration;
import org.eclipse.jdt.internal.compiler.ast.Argument;
import org.eclipse.jdt.internal.compiler.ast.Block;
import org.eclipse.jdt.internal.compiler.ast.CompilationUnitDeclaration;
import org.eclipse.jdt.internal.compiler.ast.Expression;
import org.eclipse.jdt.internal.compiler.ast.ExpressionContext;
import org.eclipse.jdt.internal.compiler.ast.FunctionalExpression;
import org.eclipse.jdt.internal.compiler.ast.IPolyExpression;
import org.eclipse.jdt.internal.compiler.ast.ReturnStatement;
import org.eclipse.jdt.internal.compiler.ast.Statement;
import org.eclipse.jdt.internal.compiler.codegen.CodeStream;
import org.eclipse.jdt.internal.compiler.env.ICompilationUnit;
import org.eclipse.jdt.internal.compiler.flow.ExceptionHandlingFlowContext;
import org.eclipse.jdt.internal.compiler.flow.ExceptionInferenceFlowContext;
import org.eclipse.jdt.internal.compiler.flow.FlowContext;
import org.eclipse.jdt.internal.compiler.flow.FlowInfo;
import org.eclipse.jdt.internal.compiler.flow.UnconditionalFlowInfo;
import org.eclipse.jdt.internal.compiler.impl.Constant;
import org.eclipse.jdt.internal.compiler.impl.ReferenceContext;
import org.eclipse.jdt.internal.compiler.lookup.AnnotationBinding;
import org.eclipse.jdt.internal.compiler.lookup.Binding;
import org.eclipse.jdt.internal.compiler.lookup.BlockScope;
import org.eclipse.jdt.internal.compiler.lookup.ClassScope;
import org.eclipse.jdt.internal.compiler.lookup.InferenceContext18;
import org.eclipse.jdt.internal.compiler.lookup.IntersectionTypeBinding18;
import org.eclipse.jdt.internal.compiler.lookup.LocalVariableBinding;
import org.eclipse.jdt.internal.compiler.lookup.LookupEnvironment;
import org.eclipse.jdt.internal.compiler.lookup.MethodBinding;
import org.eclipse.jdt.internal.compiler.lookup.MethodScope;
import org.eclipse.jdt.internal.compiler.lookup.ParameterizedTypeBinding;
import org.eclipse.jdt.internal.compiler.lookup.PolyTypeBinding;
import org.eclipse.jdt.internal.compiler.lookup.ProblemMethodBinding;
import org.eclipse.jdt.internal.compiler.lookup.ReferenceBinding;
import org.eclipse.jdt.internal.compiler.lookup.Scope;
import org.eclipse.jdt.internal.compiler.lookup.SourceTypeBinding;
import org.eclipse.jdt.internal.compiler.lookup.SyntheticArgumentBinding;
import org.eclipse.jdt.internal.compiler.lookup.TypeBinding;
import org.eclipse.jdt.internal.compiler.lookup.TypeConstants;
import org.eclipse.jdt.internal.compiler.lookup.VariableBinding;
import org.eclipse.jdt.internal.compiler.parser.Parser;
import org.eclipse.jdt.internal.compiler.problem.AbortCompilation;
import org.eclipse.jdt.internal.compiler.problem.AbortCompilationUnit;
import org.eclipse.jdt.internal.compiler.problem.AbortMethod;
import org.eclipse.jdt.internal.compiler.problem.AbortType;
import org.eclipse.jdt.internal.compiler.problem.ProblemSeverities;

public class LambdaExpression extends FunctionalExpression implements IPolyExpression, ReferenceContext, ProblemSeverities {
  public Argument[] arguments;
  
  private TypeBinding[] argumentTypes;
  
  public int arrowPosition;
  
  public Statement body;
  
  public boolean hasParentheses;
  
  public MethodScope scope;
  
  boolean voidCompatible = true;
  
  boolean valueCompatible = false;
  
  boolean returnsValue;
  
  private boolean requiresGenericSignature;
  
  boolean returnsVoid;
  
  public org.eclipse.jdt.internal.compiler.ast.LambdaExpression original = this;
  
  public SyntheticArgumentBinding[] outerLocalVariables = NO_SYNTHETIC_ARGUMENTS;
  
  private int outerLocalVariablesSlotSize = 0;
  
  private boolean assistNode = false;
  
  private boolean hasIgnoredMandatoryErrors = false;
  
  private ReferenceBinding classType;
  
  public int ordinal;
  
  private Set thrownExceptions;
  
  public char[] text;
  
  private static final SyntheticArgumentBinding[] NO_SYNTHETIC_ARGUMENTS = new SyntheticArgumentBinding[0];
  
  private static final Block NO_BODY = new Block(0);
  
  private HashMap<TypeBinding, org.eclipse.jdt.internal.compiler.ast.LambdaExpression> copiesPerTargetType;
  
  protected Expression[] resultExpressions = NO_EXPRESSIONS;
  
  public InferenceContext18 inferenceContext;
  
  public LambdaExpression(CompilationResult compilationResult, boolean assistNode, boolean requiresGenericSignature) {
    super(compilationResult);
    this.assistNode = assistNode;
    this.requiresGenericSignature = requiresGenericSignature;
    setArguments(NO_ARGUMENTS);
    setBody((Statement)NO_BODY);
  }
  
  public LambdaExpression(CompilationResult compilationResult, boolean assistNode) {
    this(compilationResult, assistNode, false);
  }
  
  public void setArguments(Argument[] arguments) {
    this.arguments = (arguments != null) ? arguments : ASTNode.NO_ARGUMENTS;
    this.argumentTypes = new TypeBinding[(arguments != null) ? arguments.length : 0];
  }
  
  public Argument[] arguments() {
    return this.arguments;
  }
  
  public TypeBinding[] argumentTypes() {
    return this.argumentTypes;
  }
  
  public void setBody(Statement body) {
    this.body = (body == null) ? (Statement)NO_BODY : body;
  }
  
  public Statement body() {
    return this.body;
  }
  
  public Expression[] resultExpressions() {
    return this.resultExpressions;
  }
  
  public void setArrowPosition(int arrowPosition) {
    this.arrowPosition = arrowPosition;
  }
  
  public int arrowPosition() {
    return this.arrowPosition;
  }
  
  protected FunctionalExpression original() {
    return this.original;
  }
  
  public void generateCode(BlockScope currentScope, CodeStream codeStream, boolean valueRequired) {
    if (this.shouldCaptureInstance) {
      this.binding.modifiers &= 0xFFFFFFF7;
    } else {
      this.binding.modifiers |= 0x8;
    } 
    SourceTypeBinding sourceType = currentScope.enclosingSourceType();
    boolean firstSpill = !(this.binding instanceof org.eclipse.jdt.internal.compiler.lookup.SyntheticMethodBinding);
    this.binding = (MethodBinding)sourceType.addSyntheticMethod(this);
    int pc = codeStream.position;
    StringBuffer signature = new StringBuffer();
    signature.append('(');
    if (this.shouldCaptureInstance) {
      codeStream.aload_0();
      signature.append(sourceType.signature());
    } 
    for (int i = 0, length = (this.outerLocalVariables == null) ? 0 : this.outerLocalVariables.length; i < length; i++) {
      SyntheticArgumentBinding syntheticArgument = this.outerLocalVariables[i];
      if (this.shouldCaptureInstance && firstSpill)
        ((LocalVariableBinding)syntheticArgument).resolvedPosition++; 
      signature.append(syntheticArgument.type.signature());
      LocalVariableBinding capturedOuterLocal = syntheticArgument.actualOuterLocalVariable;
      VariableBinding[] path = currentScope.getEmulationPath(capturedOuterLocal);
      codeStream.generateOuterAccess((Object[])path, (ASTNode)this, (Binding)capturedOuterLocal, (Scope)currentScope);
    } 
    signature.append(')');
    if (this.expectedType instanceof IntersectionTypeBinding18) {
      signature.append(((IntersectionTypeBinding18)this.expectedType).getSAMType((Scope)currentScope).signature());
    } else {
      signature.append(this.expectedType.signature());
    } 
    int invokeDynamicNumber = codeStream.classFile.recordBootstrapMethod(this);
    codeStream.invokeDynamic(invokeDynamicNumber, (this.shouldCaptureInstance ? 1 : 0) + this.outerLocalVariablesSlotSize, 1, this.descriptor.selector, signature.toString().toCharArray());
    if (!valueRequired)
      codeStream.pop(); 
    codeStream.recordPositionsFrom(pc, this.sourceStart);
  }
  
  public boolean kosherDescriptor(Scope currentScope, MethodBinding sam, boolean shouldChatter) {
    if (sam.typeVariables != Binding.NO_TYPE_VARIABLES) {
      if (shouldChatter)
        currentScope.problemReporter().lambdaExpressionCannotImplementGenericMethod(this, sam); 
      return false;
    } 
    return super.kosherDescriptor(currentScope, sam, shouldChatter);
  }
  
  public TypeBinding resolveType(BlockScope blockScope, boolean skipKosherCheck) {
    boolean argumentsTypeElided = argumentsTypeElided();
    int argumentsLength = (this.arguments == null) ? 0 : this.arguments.length;
    if (this.constant != Constant.NotAConstant) {
      this.constant = Constant.NotAConstant;
      this.enclosingScope = blockScope;
      if (this.original == this)
        this.ordinal = recordFunctionalType((Scope)blockScope); 
      if (!argumentsTypeElided)
        for (int m = 0; m < argumentsLength; m++)
          this.argumentTypes[m] = (this.arguments[m]).type.resolveType(blockScope, true);  
      if (this.expectedType == null && this.expressionContext == ExpressionContext.INVOCATION_CONTEXT)
        return (TypeBinding)new PolyTypeBinding((Expression)this); 
    } 
    MethodScope methodScope = blockScope.methodScope();
    this.scope = new MethodScope((Scope)blockScope, this, methodScope.isStatic, methodScope.lastVisibleFieldID);
    this.scope.isConstructorCall = methodScope.isConstructorCall;
    super.resolveType(blockScope, skipKosherCheck);
    boolean haveDescriptor = (this.descriptor != null);
    if (!skipKosherCheck && (!haveDescriptor || this.descriptor.typeVariables != Binding.NO_TYPE_VARIABLES))
      return this.resolvedType = null; 
    this.binding = new MethodBinding(33558530, 
        CharOperation.concat(TypeConstants.ANONYMOUS_METHOD, Integer.toString(this.ordinal).toCharArray()), 
        haveDescriptor ? this.descriptor.returnType : (TypeBinding)TypeBinding.VOID, 
        Binding.NO_PARAMETERS, 
        haveDescriptor ? this.descriptor.thrownExceptions : Binding.NO_EXCEPTIONS, 
        (ReferenceBinding)blockScope.enclosingSourceType());
    this.binding.typeVariables = Binding.NO_TYPE_VARIABLES;
    boolean argumentsHaveErrors = false;
    if (haveDescriptor) {
      int parametersLength = this.descriptor.parameters.length;
      if (parametersLength != argumentsLength) {
        this.scope.problemReporter().lambdaSignatureMismatched(this);
        if (argumentsTypeElided || this.original != this)
          return this.resolvedType = null; 
        this.resolvedType = null;
        argumentsHaveErrors = true;
      } 
    } 
    TypeBinding[] newParameters = new TypeBinding[argumentsLength];
    AnnotationBinding[][] parameterAnnotations = null;
    for (int i = 0; i < argumentsLength; i++) {
      Argument argument = this.arguments[i];
      if (argument.isVarArgs())
        if (i == argumentsLength - 1) {
          this.binding.modifiers |= 0x80;
        } else {
          this.scope.problemReporter().illegalVarargInLambda(argument);
          argumentsHaveErrors = true;
        }  
      TypeBinding expectedParameterType = (haveDescriptor && i < this.descriptor.parameters.length) ? this.descriptor.parameters[i] : null;
      TypeBinding argumentType = argumentsTypeElided ? expectedParameterType : this.argumentTypes[i];
      if (argumentType == null) {
        argumentsHaveErrors = true;
      } else if (argumentType == TypeBinding.VOID) {
        this.scope.problemReporter().argumentTypeCannotBeVoid((ASTNode)this, argument);
        argumentsHaveErrors = true;
      } else {
        if (!argumentType.isValidBinding())
          this.binding.tagBits |= 0x200L; 
        if ((argumentType.tagBits & 0x80L) != 0L)
          this.binding.tagBits |= 0x80L; 
      } 
    } 
    if (!argumentsTypeElided && !argumentsHaveErrors) {
      ReferenceBinding groundType = null;
      ReferenceBinding expectedSAMType = null;
      if (this.expectedType instanceof IntersectionTypeBinding18) {
        expectedSAMType = (ReferenceBinding)((IntersectionTypeBinding18)this.expectedType).getSAMType((Scope)blockScope);
      } else if (this.expectedType instanceof ReferenceBinding) {
        expectedSAMType = (ReferenceBinding)this.expectedType;
      } 
      if (expectedSAMType != null)
        groundType = findGroundTargetType(blockScope, (TypeBinding)expectedSAMType, argumentsTypeElided); 
      if (groundType != null) {
        this.descriptor = groundType.getSingleAbstractMethod((Scope)blockScope, true);
        if (!this.descriptor.isValidBinding()) {
          reportSamProblem(blockScope, this.descriptor);
        } else {
          if (groundType != expectedSAMType && 
            !groundType.isCompatibleWith((TypeBinding)expectedSAMType, (Scope)this.scope)) {
            blockScope.problemReporter().typeMismatchError((TypeBinding)groundType, this.expectedType, (ASTNode)this, null);
            return this.resolvedType = null;
          } 
          this.resolvedType = (TypeBinding)groundType;
        } 
      } 
    } 
    boolean genericSignatureNeeded = !(!this.requiresGenericSignature && !(blockScope.compilerOptions()).generateGenericSignatureForLambdaExpressions);
    for (int j = 0; j < argumentsLength; j++) {
      Argument argument = this.arguments[j];
      TypeBinding expectedParameterType = (haveDescriptor && j < this.descriptor.parameters.length) ? this.descriptor.parameters[j] : null;
      TypeBinding argumentType = argumentsTypeElided ? expectedParameterType : this.argumentTypes[j];
      if (argumentType != null && argumentType != TypeBinding.VOID) {
        if (haveDescriptor && expectedParameterType != null && argumentType.isValidBinding() && TypeBinding.notEquals(argumentType, expectedParameterType) && 
          expectedParameterType.isProperType(true)) {
          this.scope.problemReporter().lambdaParameterTypeMismatched(argument, argument.type, expectedParameterType);
          this.resolvedType = null;
        } 
        if (genericSignatureNeeded) {
          TypeBinding leafType = argumentType.leafComponentType();
          if (leafType instanceof ReferenceBinding && (((ReferenceBinding)leafType).modifiers & 0x40000000) != 0)
            this.binding.modifiers |= 0x40000000; 
        } 
        newParameters[j] = argument.bind(this.scope, argumentType, false);
        if (argument.annotations != null) {
          this.binding.tagBits |= 0x400L;
          if (parameterAnnotations == null) {
            parameterAnnotations = new AnnotationBinding[argumentsLength][];
            for (int m = 0; m < j; m++)
              parameterAnnotations[m] = Binding.NO_ANNOTATIONS; 
          } 
          parameterAnnotations[j] = argument.binding.getAnnotations();
        } else if (parameterAnnotations != null) {
          parameterAnnotations[j] = Binding.NO_ANNOTATIONS;
        } 
      } 
    } 
    if (!argumentsHaveErrors) {
      this.binding.parameters = newParameters;
      if (parameterAnnotations != null)
        this.binding.setParameterAnnotations(parameterAnnotations); 
    } 
    if (!argumentsTypeElided && !argumentsHaveErrors && this.binding.isVarargs() && 
      !this.binding.parameters[this.binding.parameters.length - 1].isReifiable())
      this.scope.problemReporter().possibleHeapPollutionFromVararg((AbstractVariableDeclaration)this.arguments[this.arguments.length - 1]); 
    ReferenceBinding[] exceptions = this.binding.thrownExceptions;
    int exceptionsLength = exceptions.length;
    for (int k = 0; k < exceptionsLength; k++) {
      ReferenceBinding exception = exceptions[k];
      if ((exception.tagBits & 0x80L) != 0L)
        this.binding.tagBits |= 0x80L; 
      if (genericSignatureNeeded)
        this.binding.modifiers |= exception.modifiers & 0x40000000; 
    } 
    TypeBinding returnType = this.binding.returnType;
    if (returnType != null) {
      if ((returnType.tagBits & 0x80L) != 0L)
        this.binding.tagBits |= 0x80L; 
      if (genericSignatureNeeded) {
        TypeBinding leafType = returnType.leafComponentType();
        if (leafType instanceof ReferenceBinding && (((ReferenceBinding)leafType).modifiers & 0x40000000) != 0)
          this.binding.modifiers |= 0x40000000; 
      } 
    } 
    if (haveDescriptor && !argumentsHaveErrors && (blockScope.compilerOptions()).isAnnotationBasedNullAnalysisEnabled) {
      if (!argumentsTypeElided) {
        AbstractMethodDeclaration.createArgumentBindings(this.arguments, this.binding, this.scope);
        mergeParameterNullAnnotations(blockScope);
      } 
      this.binding.tagBits |= this.descriptor.tagBits & 0x180000000000000L;
    } 
    this.binding.modifiers &= 0xFDFFFFFF;
    if (this.body instanceof Expression) {
      Expression expression = (Expression)this.body;
      (new ReturnStatement(expression, expression.sourceStart, expression.sourceEnd, true)).resolve((BlockScope)this.scope);
      if (expression.resolvedType == TypeBinding.VOID && !expression.statementExpression())
        this.scope.problemReporter().invalidExpressionAsStatement(expression); 
    } else {
      this.body.resolve((BlockScope)this.scope);
      if (!this.returnsVoid && !this.returnsValue)
        this.valueCompatible = this.body.doesNotCompleteNormally(); 
    } 
    if ((this.binding.tagBits & 0x80L) != 0L)
      this.scope.problemReporter().missingTypeInLambda(this, this.binding); 
    if (this.shouldCaptureInstance && this.scope.isConstructorCall)
      this.scope.problemReporter().fieldsOrThisBeforeConstructorInvocation((ASTNode)this); 
    return argumentsHaveErrors ? (this.resolvedType = null) : this.resolvedType;
  }
  
  private ReferenceBinding findGroundTargetType(BlockScope blockScope, TypeBinding targetType, boolean argumentTypesElided) {
    if (targetType instanceof IntersectionTypeBinding18)
      targetType = ((IntersectionTypeBinding18)targetType).getSAMType((Scope)blockScope); 
    if (targetType instanceof ReferenceBinding && targetType.isValidBinding()) {
      ParameterizedTypeBinding withWildCards = InferenceContext18.parameterizedWithWildcard(targetType);
      if (withWildCards != null) {
        if (!argumentTypesElided) {
          InferenceContext18 freshInferenceContext = new InferenceContext18((Scope)blockScope);
          try {
            return freshInferenceContext.inferFunctionalInterfaceParameterization(this, blockScope, withWildCards);
          } finally {
            freshInferenceContext.cleanUp();
          } 
        } 
        return findGroundTargetTypeForElidedLambda(blockScope, withWildCards);
      } 
      return (ReferenceBinding)targetType;
    } 
    return null;
  }
  
  public ReferenceBinding findGroundTargetTypeForElidedLambda(BlockScope blockScope, ParameterizedTypeBinding withWildCards) {
    TypeBinding[] types = withWildCards.getNonWildcardParameterization((Scope)blockScope);
    if (types == null)
      return null; 
    ReferenceBinding genericType = withWildCards.genericType();
    return (ReferenceBinding)blockScope.environment().createParameterizedType(genericType, types, withWildCards.enclosingType());
  }
  
  public boolean argumentsTypeElided() {
    return (this.arguments.length > 0 && this.arguments[0].hasElidedType());
  }
  
  private void analyzeExceptions() {
    try {
      ExceptionInferenceFlowContext exceptionInferenceFlowContext;
      this.body.analyseCode((BlockScope)this.scope, 
          (FlowContext)(exceptionInferenceFlowContext = new ExceptionInferenceFlowContext(null, (ASTNode)this, Binding.NO_EXCEPTIONS, null, (BlockScope)this.scope, FlowInfo.DEAD_END)), 
          (FlowInfo)UnconditionalFlowInfo.fakeInitializedFlowInfo((this.scope.outerMostMethodScope()).analysisIndex, (this.scope.referenceType()).maxFieldCount));
      this.thrownExceptions = (((ExceptionHandlingFlowContext)exceptionInferenceFlowContext).extendedExceptions == null) ? Collections.emptySet() : new HashSet(((ExceptionHandlingFlowContext)exceptionInferenceFlowContext).extendedExceptions);
    } catch (Exception exception) {}
  }
  
  public FlowInfo analyseCode(BlockScope currentScope, FlowContext flowContext, FlowInfo flowInfo) {
    if (this.ignoreFurtherInvestigation)
      return flowInfo; 
    FlowInfo lambdaInfo = flowInfo.copy();
    ExceptionHandlingFlowContext methodContext = 
      new ExceptionHandlingFlowContext(
        flowContext, 
        (ASTNode)this, 
        this.binding.thrownExceptions, 
        flowContext.getInitializationContext(), 
        (BlockScope)this.scope, 
        FlowInfo.DEAD_END);
    MethodBinding methodWithParameterDeclaration = argumentsTypeElided() ? this.descriptor : this.binding;
    AbstractMethodDeclaration.analyseArguments(currentScope.environment(), lambdaInfo, this.arguments, methodWithParameterDeclaration);
    if (this.arguments != null)
      for (int i = 0, count = this.arguments.length; i < count; i++)
        this.bits |= (this.arguments[i]).bits & 0x100000;  
    lambdaInfo = this.body.analyseCode((BlockScope)this.scope, (FlowContext)methodContext, lambdaInfo);
    if (this.body instanceof Block) {
      TypeBinding returnTypeBinding = expectedResultType();
      if (returnTypeBinding == TypeBinding.VOID) {
        if ((lambdaInfo.tagBits & 0x1) == 0 || ((Block)this.body).statements == null)
          this.bits |= 0x40; 
      } else if (lambdaInfo != FlowInfo.DEAD_END) {
        this.scope.problemReporter().shouldReturn(returnTypeBinding, (ASTNode)this);
      } 
    } else if ((currentScope.compilerOptions()).isAnnotationBasedNullAnalysisEnabled && 
      lambdaInfo.reachMode() == 0) {
      Expression expression = (Expression)this.body;
      checkAgainstNullAnnotation(flowContext, expression, flowInfo, expression.nullStatus(lambdaInfo, flowContext));
    } 
    return flowInfo;
  }
  
  void validateNullAnnotations() {
    if (this.binding != null) {
      int length = this.binding.parameters.length;
      for (int i = 0; i < length; i++) {
        if (!this.scope.validateNullAnnotation(this.binding.returnType.tagBits, (this.arguments[i]).type, (this.arguments[i]).annotations))
          this.binding.returnType = this.binding.returnType.withoutToplevelNullAnnotation(); 
      } 
    } 
  }
  
  private void mergeParameterNullAnnotations(BlockScope currentScope) {
    LookupEnvironment env = currentScope.environment();
    TypeBinding[] ourParameters = this.binding.parameters;
    TypeBinding[] descParameters = this.descriptor.parameters;
    int len = Math.min(ourParameters.length, descParameters.length);
    for (int i = 0; i < len; i++) {
      long ourTagBits = (ourParameters[i]).tagBits & 0x180000000000000L;
      long descTagBits = (descParameters[i]).tagBits & 0x180000000000000L;
      if (ourTagBits == 0L) {
        if (descTagBits != 0L && !ourParameters[i].isBaseType()) {
          AnnotationBinding[] annotations = descParameters[i].getTypeAnnotations();
          for (int j = 0, length = annotations.length; j < length; j++) {
            AnnotationBinding annotation = annotations[j];
            if (annotation != null && annotation.getAnnotationType().hasNullBit(96))
              ourParameters[i] = env.createAnnotatedType(ourParameters[i], new AnnotationBinding[] { annotation }); 
          } 
        } 
      } else if (ourTagBits != descTagBits && 
        ourTagBits == 72057594037927936L) {
        char[][] inheritedAnnotationName = null;
        if (descTagBits == 36028797018963968L)
          inheritedAnnotationName = env.getNullableAnnotationName(); 
        currentScope.problemReporter().illegalRedefinitionToNonNullParameter(this.arguments[i], this.descriptor.declaringClass, inheritedAnnotationName);
      } 
    } 
  }
  
  void checkAgainstNullAnnotation(FlowContext flowContext, Expression expression, FlowInfo flowInfo, int nullStatus) {
    if (nullStatus != 4)
      if ((this.descriptor.returnType.tagBits & 0x100000000000000L) != 0L)
        flowContext.recordNullityMismatch((BlockScope)this.scope, expression, expression.resolvedType, this.descriptor.returnType, flowInfo, nullStatus, null);  
  }
  
  public boolean isPertinentToApplicability(TypeBinding targetType, MethodBinding method) {
    if (targetType == null)
      return true; 
    if (argumentsTypeElided())
      return false; 
    if (!super.isPertinentToApplicability(targetType, method))
      return false; 
    if (this.body instanceof Expression) {
      if (!((Expression)this.body).isPertinentToApplicability(targetType, method))
        return false; 
    } else {
      Expression[] returnExpressions = this.resultExpressions;
      if (returnExpressions != NO_EXPRESSIONS) {
        for (int i = 0, length = returnExpressions.length; i < length; i++) {
          if (!returnExpressions[i].isPertinentToApplicability(targetType, method))
            return false; 
        } 
      } else {
        try {
          this.body.traverse((ASTVisitor)new ResultsAnalyser(this, targetType, method), (BlockScope)this.scope);
        } catch (null notPertientToApplicability) {
          return false;
        } 
      } 
    } 
    return true;
  }
  
  public boolean isVoidCompatible() {
    return this.voidCompatible;
  }
  
  public boolean isValueCompatible() {
    return this.valueCompatible;
  }
  
  public StringBuffer printExpression(int tab, StringBuffer output) {
    return printExpression(tab, output, false);
  }
  
  public StringBuffer printExpression(int tab, StringBuffer output, boolean makeShort) {
    int parenthesesCount = (this.bits & 0x1FE00000) >> 21;
    String suffix = "";
    int i;
    for (i = 0; i < parenthesesCount; i++) {
      output.append('(');
      suffix = String.valueOf(suffix) + ')';
    } 
    output.append('(');
    if (this.arguments != null)
      for (i = 0; i < this.arguments.length; i++) {
        if (i > 0)
          output.append(", "); 
        this.arguments[i].print(0, output);
      }  
    output.append(") -> ");
    if (makeShort) {
      output.append("{}");
    } else if (this.body != null) {
      this.body.print((this.body instanceof Block) ? tab : 0, output);
    } else {
      output.append("<@incubator>");
    } 
    return output.append(suffix);
  }
  
  public TypeBinding expectedResultType() {
    return (this.descriptor != null && this.descriptor.isValidBinding()) ? this.descriptor.returnType : null;
  }
  
  public void traverse(ASTVisitor visitor, BlockScope blockScope) {
    if (visitor.visit(this, blockScope)) {
      if (this.arguments != null) {
        int argumentsLength = this.arguments.length;
        for (int i = 0; i < argumentsLength; i++)
          this.arguments[i].traverse(visitor, (BlockScope)this.scope); 
      } 
      if (this.body != null)
        this.body.traverse(visitor, (BlockScope)this.scope); 
    } 
    visitor.endVisit(this, blockScope);
  }
  
  public MethodScope getScope() {
    return this.scope;
  }
  
  private boolean enclosingScopesHaveErrors() {
    BlockScope blockScope = this.enclosingScope;
    while (blockScope != null) {
      ReferenceContext context = blockScope.referenceContext();
      if (context != null && context.hasErrors())
        return true; 
      Scope scope = ((Scope)blockScope).parent;
    } 
    return false;
  }
  
  private void analyzeShape() {
    if (this.body instanceof Expression) {
      this.voidCompatible = this.assistNode ? true : ((Expression)this.body).statementExpression();
      this.valueCompatible = true;
    } else {
      if (this.assistNode) {
        this.voidCompatible = true;
        this.valueCompatible = true;
      } 
      this.body.traverse((ASTVisitor)new ShapeComputer(this), null);
      if (!this.returnsValue && !this.returnsVoid)
        this.valueCompatible = this.body.doesNotCompleteNormally(); 
    } 
  }
  
  public boolean isPotentiallyCompatibleWith(TypeBinding targetType, Scope skope) {
    if (!super.isPertinentToApplicability(targetType, null))
      return true; 
    MethodBinding sam = targetType.getSingleAbstractMethod(skope, true);
    if (sam == null || !sam.isValidBinding())
      return false; 
    if (sam.parameters.length != this.arguments.length)
      return false; 
    analyzeShape();
    if (sam.returnType.id == 6) {
      if (!this.voidCompatible)
        return false; 
    } else if (!this.valueCompatible) {
      return false;
    } 
    return true;
  }
  
  public boolean isCompatibleWith(TypeBinding targetType, Scope skope) {
    if (!super.isPertinentToApplicability(targetType, null))
      return true; 
    org.eclipse.jdt.internal.compiler.ast.LambdaExpression copy = null;
    try {
      copy = cachedResolvedCopy(targetType, argumentsTypeElided(), false, null);
    } catch (org.eclipse.jdt.internal.compiler.ast.LambdaExpression.CopyFailureException copyFailureException) {
      if (this.assistNode)
        return true; 
      return !isPertinentToApplicability(targetType, null);
    } 
    if (copy == null)
      return false; 
    ReferenceBinding referenceBinding = findGroundTargetType(this.enclosingScope, targetType, argumentsTypeElided());
    MethodBinding sam = referenceBinding.getSingleAbstractMethod((Scope)this.enclosingScope, true);
    if (sam == null || sam.problemId() == 17)
      return false; 
    if (sam.returnType.id == 6) {
      if (!copy.voidCompatible)
        return false; 
    } else if (!copy.valueCompatible) {
      return false;
    } 
    if (!isPertinentToApplicability((TypeBinding)referenceBinding, null))
      return true; 
    if (!kosherDescriptor((Scope)this.enclosingScope, sam, false))
      return false; 
    Expression[] returnExpressions = copy.resultExpressions;
    for (int i = 0, length = returnExpressions.length; i < length; i++) {
      if (this.enclosingScope.parameterCompatibilityLevel((returnExpressions[i]).resolvedType, sam.returnType) == -1 && 
        !returnExpressions[i].isConstantValueOfTypeAssignableToType((returnExpressions[i]).resolvedType, sam.returnType) && (
        sam.returnType.id != 6 || this.body instanceof Block))
        return false; 
    } 
    return true;
  }
  
  private org.eclipse.jdt.internal.compiler.ast.LambdaExpression cachedResolvedCopy(TypeBinding targetType, boolean anyTargetOk, boolean requireExceptionAnalysis, InferenceContext18 context) {
    ReferenceBinding referenceBinding = findGroundTargetType(this.enclosingScope, targetType, argumentsTypeElided());
    if (referenceBinding == null)
      return null; 
    MethodBinding sam = referenceBinding.getSingleAbstractMethod((Scope)this.enclosingScope, true);
    if (sam == null || !sam.isValidBinding())
      return null; 
    if (sam.parameters.length != this.arguments.length)
      return null; 
    org.eclipse.jdt.internal.compiler.ast.LambdaExpression copy = null;
    if (this.copiesPerTargetType != null) {
      copy = this.copiesPerTargetType.get(referenceBinding);
      if (copy == null && 
        anyTargetOk && this.copiesPerTargetType.values().size() > 0)
        copy = this.copiesPerTargetType.values().iterator().next(); 
    } 
    IErrorHandlingPolicy oldPolicy = this.enclosingScope.problemReporter().switchErrorHandlingPolicy(silentErrorHandlingPolicy);
    try {
      if (copy == null) {
        copy = copy();
        if (copy == null)
          throw new CopyFailureException(this); 
        copy.setExpressionContext(this.expressionContext);
        copy.setExpectedType((TypeBinding)referenceBinding);
        copy.inferenceContext = context;
        TypeBinding type = copy.resolveType(this.enclosingScope, true);
        if (type == null || !type.isValidBinding())
          return null; 
        if (this.copiesPerTargetType == null)
          this.copiesPerTargetType = new HashMap<>(); 
        this.copiesPerTargetType.put(referenceBinding, copy);
      } 
      if (!requireExceptionAnalysis)
        return copy; 
      if (copy.thrownExceptions == null && 
        !copy.hasIgnoredMandatoryErrors && !enclosingScopesHaveErrors())
        copy.analyzeExceptions(); 
      return copy;
    } finally {
      this.enclosingScope.problemReporter().switchErrorHandlingPolicy(oldPolicy);
    } 
  }
  
  public org.eclipse.jdt.internal.compiler.ast.LambdaExpression resolveExpressionExpecting(TypeBinding targetType, Scope skope, InferenceContext18 context) {
    org.eclipse.jdt.internal.compiler.ast.LambdaExpression copy = null;
    try {
      copy = cachedResolvedCopy(targetType, false, true, context);
    } catch (org.eclipse.jdt.internal.compiler.ast.LambdaExpression.CopyFailureException copyFailureException) {
      return null;
    } 
    return copy;
  }
  
  public boolean sIsMoreSpecific(TypeBinding s, TypeBinding t, Scope skope) {
    if (super.sIsMoreSpecific(s, t, skope))
      return true; 
    if (argumentsTypeElided() || t.findSuperTypeOriginatingFrom(s) != null)
      return false; 
    s = s.capture((Scope)this.enclosingScope, this.sourceStart, this.sourceEnd);
    MethodBinding sSam = s.getSingleAbstractMethod((Scope)this.enclosingScope, true);
    if (sSam == null || !sSam.isValidBinding())
      return false; 
    TypeBinding r1 = sSam.returnType;
    MethodBinding tSam = t.getSingleAbstractMethod((Scope)this.enclosingScope, true);
    if (tSam == null || !tSam.isValidBinding())
      return true; 
    TypeBinding r2 = tSam.returnType;
    if (r2.id == 6)
      return true; 
    if (r1.id == 6)
      return false; 
    if (r1.isCompatibleWith(r2, skope))
      return true; 
    org.eclipse.jdt.internal.compiler.ast.LambdaExpression copy = cachedResolvedCopy(s, true, false, null);
    Expression[] returnExpressions = copy.resultExpressions;
    int returnExpressionsLength = (returnExpressions == null) ? 0 : returnExpressions.length;
    if (returnExpressionsLength > 0) {
      if (r1.isBaseType() && !r2.isBaseType()) {
        int i;
        for (i = 0; i < returnExpressionsLength && 
          !returnExpressions[i].isPolyExpression() && (returnExpressions[i]).resolvedType.isBaseType(); i++);
        if (i == returnExpressionsLength)
          return true; 
      } 
      if (!r1.isBaseType() && r2.isBaseType()) {
        int i;
        for (i = 0; i < returnExpressionsLength && 
          !(returnExpressions[i]).resolvedType.isBaseType(); i++);
        if (i == returnExpressionsLength)
          return true; 
      } 
      if (r1.isFunctionalInterface((Scope)this.enclosingScope) && r2.isFunctionalInterface((Scope)this.enclosingScope)) {
        int i;
        for (i = 0; i < returnExpressionsLength; i++) {
          Expression resultExpression = returnExpressions[i];
          if (!resultExpression.sIsMoreSpecific(r1, r2, skope))
            break; 
        } 
        if (i == returnExpressionsLength)
          return true; 
      } 
    } 
    return false;
  }
  
  org.eclipse.jdt.internal.compiler.ast.LambdaExpression copy() {
    Parser parser = new Parser(this.enclosingScope.problemReporter(), false);
    ICompilationUnit compilationUnit = this.compilationResult.getCompilationUnit();
    char[] source = (compilationUnit != null) ? compilationUnit.getContents() : this.text;
    org.eclipse.jdt.internal.compiler.ast.LambdaExpression copy = (org.eclipse.jdt.internal.compiler.ast.LambdaExpression)parser.parseLambdaExpression(source, (compilationUnit != null) ? this.sourceStart : 0, this.sourceEnd - this.sourceStart + 1, 
        this.enclosingScope.referenceCompilationUnit(), false);
    if (copy != null) {
      copy.original = this;
      copy.assistNode = this.assistNode;
      copy.enclosingScope = this.enclosingScope;
    } 
    return copy;
  }
  
  public void returnsExpression(Expression expression, TypeBinding resultType) {
    if (this.original == this)
      return; 
    if (this.body instanceof Expression) {
      this.valueCompatible = !(resultType != null && resultType.id == 6);
      this.voidCompatible = this.assistNode ? true : ((Expression)this.body).statementExpression();
      this.resultExpressions = new Expression[] { expression };
      return;
    } 
    if (expression != null) {
      this.returnsValue = true;
      this.voidCompatible = false;
      this.valueCompatible = !this.returnsVoid;
      Expression[] returnExpressions = this.resultExpressions;
      int resultsLength = returnExpressions.length;
      System.arraycopy(returnExpressions, 0, returnExpressions = new Expression[resultsLength + 1], 0, resultsLength);
      returnExpressions[resultsLength] = expression;
      this.resultExpressions = returnExpressions;
    } else {
      this.returnsVoid = true;
      this.valueCompatible = false;
      this.voidCompatible = !this.returnsValue;
    } 
  }
  
  public CompilationResult compilationResult() {
    return this.compilationResult;
  }
  
  public void abort(int abortLevel, CategorizedProblem problem) {
    switch (abortLevel) {
      case 2:
        throw new AbortCompilation(this.compilationResult, problem);
      case 4:
        throw new AbortCompilationUnit(this.compilationResult, problem);
      case 8:
        throw new AbortType(this.compilationResult, problem);
    } 
    throw new AbortMethod(this.compilationResult, problem);
  }
  
  public CompilationUnitDeclaration getCompilationUnitDeclaration() {
    return (this.enclosingScope == null) ? null : (this.enclosingScope.compilationUnitScope()).referenceContext;
  }
  
  public boolean hasErrors() {
    return this.ignoreFurtherInvestigation;
  }
  
  public void tagAsHavingErrors() {
    this.ignoreFurtherInvestigation = true;
    Scope parent = this.enclosingScope.parent;
    while (parent != null) {
      ReferenceContext parentAST;
      switch (parent.kind) {
        case 2:
        case 3:
          parentAST = parent.referenceContext();
          if (parentAST != this) {
            parentAST.tagAsHavingErrors();
            return;
          } 
          break;
      } 
      parent = parent.parent;
    } 
  }
  
  public void tagAsHavingIgnoredMandatoryErrors(int problemId) {
    switch (problemId) {
      case 16777362:
      case 16777384:
      case 16778098:
        return;
      case 99:
      case 16777235:
      case 67108969:
      case 67109635:
      case 553648781:
      case 553648783:
      case 553648784:
      case 553648785:
      case 553648786:
      case 553648787:
      case 603979884:
        return;
    } 
    this.hasIgnoredMandatoryErrors = true;
    MethodScope enclosingLambdaScope = (this.scope == null) ? null : this.scope.enclosingLambdaScope();
    while (enclosingLambdaScope != null) {
      org.eclipse.jdt.internal.compiler.ast.LambdaExpression enclosingLambda = (org.eclipse.jdt.internal.compiler.ast.LambdaExpression)enclosingLambdaScope.referenceContext;
      enclosingLambda.hasIgnoredMandatoryErrors = true;
      enclosingLambdaScope = enclosingLambdaScope.enclosingLambdaScope();
    } 
  }
  
  public Set<TypeBinding> getThrownExceptions() {
    if (this.thrownExceptions == null)
      return Collections.emptySet(); 
    return this.thrownExceptions;
  }
  
  public void generateCode(ClassScope classScope, ClassFile classFile) {
    int problemResetPC = 0;
    classFile.codeStream.wideMode = false;
    boolean restart = false;
    do {
      try {
        problemResetPC = classFile.contentsOffset;
        generateCode(classFile);
        restart = false;
      } catch (AbortMethod e) {
        if (e.compilationResult == CodeStream.RESTART_IN_WIDE_MODE) {
          classFile.contentsOffset = problemResetPC;
          classFile.methodCount--;
          classFile.codeStream.resetInWideMode();
          restart = true;
        } else if (e.compilationResult == CodeStream.RESTART_CODE_GEN_FOR_UNUSED_LOCALS_MODE) {
          classFile.contentsOffset = problemResetPC;
          classFile.methodCount--;
          classFile.codeStream.resetForCodeGenUnusedLocals();
          restart = true;
        } else {
          throw new AbortType(this.compilationResult, e.problem);
        } 
      } 
    } while (restart);
  }
  
  public void generateCode(ClassFile classFile) {
    classFile.generateMethodInfoHeader(this.binding);
    int methodAttributeOffset = classFile.contentsOffset;
    int attributeNumber = classFile.generateMethodInfoAttributes(this.binding);
    int codeAttributeOffset = classFile.contentsOffset;
    classFile.generateCodeAttributeHeader();
    CodeStream codeStream = classFile.codeStream;
    codeStream.reset(this, classFile);
    this.scope.computeLocalVariablePositions(this.outerLocalVariablesSlotSize + (this.binding.isStatic() ? 0 : 1), codeStream);
    if (this.outerLocalVariables != null)
      for (int i = 0, max = this.outerLocalVariables.length; i < max; i++) {
        SyntheticArgumentBinding syntheticArgumentBinding;
        codeStream.addVisibleLocalVariable((LocalVariableBinding)(syntheticArgumentBinding = this.outerLocalVariables[i]));
        codeStream.record((LocalVariableBinding)syntheticArgumentBinding);
        syntheticArgumentBinding.recordInitializationStartPC(0);
      }  
    if (this.arguments != null)
      for (int i = 0, max = this.arguments.length; i < max; i++) {
        LocalVariableBinding argBinding;
        codeStream.addVisibleLocalVariable(argBinding = (this.arguments[i]).binding);
        argBinding.recordInitializationStartPC(0);
      }  
    if (this.body instanceof Block) {
      this.body.generateCode((BlockScope)this.scope, codeStream);
      if ((this.bits & 0x40) != 0)
        codeStream.return_(); 
    } else {
      Expression expression = (Expression)this.body;
      expression.generateCode((BlockScope)this.scope, codeStream, true);
      if (this.binding.returnType == TypeBinding.VOID) {
        codeStream.return_();
      } else {
        codeStream.generateReturnBytecode(expression);
      } 
    } 
    codeStream.exitUserScope((BlockScope)this.scope);
    codeStream.recordPositionsFrom(0, this.sourceEnd);
    try {
      classFile.completeCodeAttribute(codeAttributeOffset);
    } catch (NegativeArraySizeException negativeArraySizeException) {
      throw new AbortMethod((this.scope.referenceCompilationUnit()).compilationResult, null);
    } 
    attributeNumber++;
    classFile.completeMethodInfo(this.binding, methodAttributeOffset, attributeNumber);
  }
  
  public void addSyntheticArgument(LocalVariableBinding actualOuterLocalVariable) {
    if (this.original != this || this.binding == null)
      return; 
    SyntheticArgumentBinding syntheticLocal = null;
    int newSlot = this.outerLocalVariables.length;
    for (int i = 0; i < newSlot; i++) {
      if ((this.outerLocalVariables[i]).actualOuterLocalVariable == actualOuterLocalVariable)
        return; 
    } 
    System.arraycopy(this.outerLocalVariables, 0, this.outerLocalVariables = new SyntheticArgumentBinding[newSlot + 1], 0, newSlot);
    this.outerLocalVariables[newSlot] = syntheticLocal = new SyntheticArgumentBinding(actualOuterLocalVariable);
    syntheticLocal.resolvedPosition = this.outerLocalVariablesSlotSize;
    syntheticLocal.declaringScope = (BlockScope)this.scope;
    int parameterCount = this.binding.parameters.length;
    TypeBinding[] newParameters = new TypeBinding[parameterCount + 1];
    newParameters[newSlot] = actualOuterLocalVariable.type;
    for (int k = 0, j = 0; k < parameterCount; k++, j++) {
      if (k == newSlot)
        j++; 
      newParameters[j] = this.binding.parameters[k];
    } 
    this.binding.parameters = newParameters;
    switch (syntheticLocal.type.id) {
      case 7:
      case 8:
        this.outerLocalVariablesSlotSize += 2;
        return;
    } 
    this.outerLocalVariablesSlotSize++;
  }
  
  public SyntheticArgumentBinding getSyntheticArgument(LocalVariableBinding actualOuterLocalVariable) {
    for (int i = 0, length = (this.outerLocalVariables == null) ? 0 : this.outerLocalVariables.length; i < length; i++) {
      if ((this.outerLocalVariables[i]).actualOuterLocalVariable == actualOuterLocalVariable)
        return this.outerLocalVariables[i]; 
    } 
    return null;
  }
  
  public MethodBinding getMethodBinding() {
    if (this.actualMethodBinding == null)
      if (this.binding != null) {
        TypeBinding[] newParams = null;
        if (this.binding instanceof org.eclipse.jdt.internal.compiler.lookup.SyntheticMethodBinding && this.outerLocalVariables.length > 0) {
          newParams = new TypeBinding[this.binding.parameters.length - this.outerLocalVariables.length];
          System.arraycopy(this.binding.parameters, this.outerLocalVariables.length, newParams, 0, newParams.length);
        } else {
          newParams = this.binding.parameters;
        } 
        this.actualMethodBinding = new MethodBinding(this.binding.modifiers, this.binding.selector, 
            this.binding.returnType, newParams, this.binding.thrownExceptions, this.binding.declaringClass);
        this.actualMethodBinding.tagBits = this.binding.tagBits;
      } else {
        this.actualMethodBinding = (MethodBinding)new ProblemMethodBinding(CharOperation.NO_CHAR, null, 17);
      }  
    return this.actualMethodBinding;
  }
  
  public int diagnosticsSourceEnd() {
    return (this.body instanceof Block) ? this.arrowPosition : this.sourceEnd;
  }
  
  public TypeBinding[] getMarkerInterfaces() {
    if (this.expectedType instanceof IntersectionTypeBinding18) {
      Set<ReferenceBinding> markerBindings = new LinkedHashSet();
      IntersectionTypeBinding18 intersectionType = (IntersectionTypeBinding18)this.expectedType;
      ReferenceBinding[] arrayOfReferenceBinding = intersectionType.intersectingTypes;
      TypeBinding samType = intersectionType.getSAMType((Scope)this.enclosingScope);
      for (int i = 0, max = arrayOfReferenceBinding.length; i < max; i++) {
        ReferenceBinding referenceBinding = arrayOfReferenceBinding[i];
        if (referenceBinding.isInterface() && 
          !TypeBinding.equalsEquals(samType, (TypeBinding)referenceBinding) && 
          ((TypeBinding)referenceBinding).id != 37)
          markerBindings.add(referenceBinding); 
      } 
      if (markerBindings.size() > 0)
        return markerBindings.<TypeBinding>toArray(new TypeBinding[markerBindings.size()]); 
    } 
    return null;
  }
  
  public ReferenceBinding getTypeBinding() {
    if (this.classType != null || this.resolvedType == null)
      return null; 
    return this.classType = (ReferenceBinding)new LambdaTypeBinding(this);
  }
}
